@@ -1,3 +1,5 @@
+// SPDX-License-Identifier: Apache-2.0
+// 
 // Copyright 2008-2016 Conrad Sanderson (http://conradsanderson.id.au)
 // Copyright 2008-2016 National ICT Australia (NICTA)
 // 
@@ -41,15 +43,80 @@ op_vectorise_col::apply_direct(Mat<typename T1::elem_type>& out, const T1& expr)
   
   typedef typename T1::elem_type eT;
   
-  if(is_same_type< T1, subview<eT> >::yes)
+  // allow detection of in-place operation
+  if(is_Mat<T1>::value || (arma_config::openmp && Proxy<T1>::use_mp))
     {
-    op_vectorise_col::apply_subview(out, reinterpret_cast< const subview<eT>& >(expr));
+    const unwrap<T1> U(expr);
+    
+    if(&out == &(U.M))
+      {
+      // output matrix is the same as the input matrix
+      
+      out.set_size(out.n_elem, 1);  // set_size() doesn't destroy data as long as the number of elements in the matrix remains the same
+      }
+    else
+      {
+      out.set_size(U.M.n_elem, 1);
+      
+      arrayops::copy(out.memptr(), U.M.memptr(), U.M.n_elem);
+      }
+    }
+  else
+  if(is_subview<T1>::value)
+    {
+    const subview<eT>& sv = reinterpret_cast< const subview<eT>& >(expr);
+    
+    if(&out == &(sv.m))
+      {
+      Mat<eT> tmp;
+      
+      op_vectorise_col::apply_subview(tmp, sv);
+      
+      out.steal_mem(tmp);
+      }
+    else
+      {
+      op_vectorise_col::apply_subview(out, sv);
+      }
     }
   else
     {
     const Proxy<T1> P(expr);
     
-    op_vectorise_col::apply_proxy(out, P);
+    const bool is_alias = P.is_alias(out);
+    
+    if(is_Mat<typename Proxy<T1>::stored_type>::value)
+      {
+      const quasi_unwrap<typename Proxy<T1>::stored_type> U(P.Q);
+      
+      if(is_alias)
+        {
+        Mat<eT> tmp(U.M.memptr(), U.M.n_elem, 1);
+        
+        out.steal_mem(tmp);
+        }
+      else
+        {
+        out.set_size(U.M.n_elem, 1);
+        
+        arrayops::copy(out.memptr(), U.M.memptr(), U.M.n_elem);
+        }
+      }
+    else
+      {
+      if(is_alias)
+        {
+        Mat<eT> tmp;
+        
+        op_vectorise_col::apply_proxy(tmp, P);
+        
+        out.steal_mem(tmp);
+        }
+      else
+        {
+        op_vectorise_col::apply_proxy(out, P);
+        }
+      }
     }
   }
 
@@ -62,31 +129,18 @@ op_vectorise_col::apply_subview(Mat<eT>& out, const subview<eT>& sv)
   {
   arma_extra_debug_sigprint();
   
-  const bool is_alias = (&out == &(sv.m));
+  const uword sv_n_rows = sv.n_rows;
+  const uword sv_n_cols = sv.n_cols;
   
-  if(is_alias == false)
+  out.set_size(sv.n_elem, 1);
+  
+  eT* out_mem = out.memptr();
+  
+  for(uword col=0; col < sv_n_cols; ++col)
     {
-    const uword sv_n_rows = sv.n_rows;
-    const uword sv_n_cols = sv.n_cols;
+    arrayops::copy(out_mem, sv.colptr(col), sv_n_rows);
     
-    out.set_size(sv.n_elem, 1);
-    
-    eT* out_mem = out.memptr();
-    
-    for(uword col=0; col < sv_n_cols; ++col)
-      {
-      arrayops::copy(out_mem, sv.colptr(col), sv_n_rows);
-      
-      out_mem += sv_n_rows;
-      }
-    }
-  else
-    {
-    Mat<eT> tmp;
-    
-    op_vectorise_col::apply_subview(tmp, sv);
-    
-    out.steal_mem(tmp);
+    out_mem += sv_n_rows;
     }
   }
 
@@ -101,83 +155,54 @@ op_vectorise_col::apply_proxy(Mat<typename T1::elem_type>& out, const Proxy<T1>&
   
   typedef typename T1::elem_type eT;
   
-  if(P.is_alias(out) == false)
+  const uword N = P.get_n_elem();
+  
+  out.set_size(N, 1);
+  
+  eT* outmem = out.memptr();
+  
+  if(Proxy<T1>::use_at == false)
     {
-    const uword N = P.get_n_elem();
+    // TODO: add handling of aligned access ?
     
-    out.set_size(N, 1);
-      
-    if(is_Mat<typename Proxy<T1>::stored_type>::value == true)
+    typename Proxy<T1>::ea_type A = P.get_ea();
+    
+    uword i,j;
+    
+    for(i=0, j=1; j < N; i+=2, j+=2)
       {
-      const unwrap<typename Proxy<T1>::stored_type> tmp(P.Q);
+      const eT tmp_i = A[i];
+      const eT tmp_j = A[j];
       
-      arrayops::copy(out.memptr(), tmp.M.memptr(), N);
+      outmem[i] = tmp_i;
+      outmem[j] = tmp_j;
       }
-    else
+    
+    if(i < N)
       {
-      eT* outmem = out.memptr();
-      
-      if(Proxy<T1>::use_at == false)
-        {
-        // TODO: add handling of aligned access ?
-        
-        typename Proxy<T1>::ea_type A = P.get_ea();
-        
-        uword i,j;
-        
-        for(i=0, j=1; j < N; i+=2, j+=2)
-          {
-          const eT tmp_i = A[i];
-          const eT tmp_j = A[j];
-          
-          outmem[i] = tmp_i;
-          outmem[j] = tmp_j;
-          }
-        
-        if(i < N)
-          {
-          outmem[i] = A[i];
-          }
-        }
-      else
-        {
-        const uword n_rows = P.get_n_rows();
-        const uword n_cols = P.get_n_cols();
-        
-        if(n_rows == 1)
-          {
-          for(uword i=0; i < n_cols; ++i)
-            {
-            outmem[i] = P.at(0,i);
-            }
-          }
-        else
-          {
-          for(uword col=0; col < n_cols; ++col)
-          for(uword row=0; row < n_rows; ++row)
-            {
-            *outmem = P.at(row,col);
-            outmem++;
-            }
-          }
-        }
+      outmem[i] = A[i];
       }
     }
-  else  // we have aliasing
+  else
     {
-    arma_extra_debug_print("op_vectorise_col::apply(): aliasing detected");
+    const uword n_rows = P.get_n_rows();
+    const uword n_cols = P.get_n_cols();
     
-    if( (is_Mat<typename Proxy<T1>::stored_type>::value == true) && (Proxy<T1>::fake_mat == false) )
+    if(n_rows == 1)
       {
-      out.set_size(out.n_elem, 1);  // set_size() doesn't destroy data as long as the number of elements in the matrix remains the same
+      for(uword i=0; i < n_cols; ++i)
+        {
+        outmem[i] = P.at(0,i);
+        }
       }
     else
       {
-      Mat<eT> tmp;
-      
-      op_vectorise_col::apply_proxy(tmp, P);
-      
-      out.steal_mem(tmp);
+      for(uword col=0; col < n_cols; ++col)
+      for(uword row=0; row < n_rows; ++row)
+        {
+        *outmem = P.at(row,col);
+        outmem++;
+        }
       }
     }
   }
@@ -203,9 +228,22 @@ op_vectorise_row::apply_direct(Mat<typename T1::elem_type>& out, const T1& expr)
   {
   arma_extra_debug_sigprint();
   
+  typedef typename T1::elem_type eT;
+  
   const Proxy<T1> P(expr);
   
-  op_vectorise_row::apply_proxy(out, P);
+  if(P.is_alias(out))
+    {
+    Mat<eT> tmp;
+    
+    op_vectorise_row::apply_proxy(tmp, P);
+    
+    out.steal_mem(tmp);
+    }
+  else
+    {
+    op_vectorise_row::apply_proxy(out, P);
+    }
   }
 
 
@@ -219,60 +257,47 @@ op_vectorise_row::apply_proxy(Mat<typename T1::elem_type>& out, const Proxy<T1>&
   
   typedef typename T1::elem_type eT;
   
-  if(P.is_alias(out) == false)
+  const uword n_rows = P.get_n_rows();
+  const uword n_cols = P.get_n_cols();
+  const uword n_elem = P.get_n_elem();
+  
+  out.set_size(1, n_elem);
+  
+  eT* outmem = out.memptr();
+  
+  if(n_cols == 1)
     {
-    const uword n_rows = P.get_n_rows();
-    const uword n_cols = P.get_n_cols();
-    const uword n_elem = P.get_n_elem();
-    
-    out.set_size(1, n_elem);
-    
-    eT* outmem = out.memptr();
-    
-    if(n_cols == 1)
+    if(is_Mat<typename Proxy<T1>::stored_type>::value)
       {
-      if(is_Mat<typename Proxy<T1>::stored_type>::value == true)
-        {
-        const unwrap<typename Proxy<T1>::stored_type> tmp(P.Q);
-        
-        arrayops::copy(out.memptr(), tmp.M.memptr(), n_elem);
-        }
-      else
-        {
-        for(uword i=0; i < n_elem; ++i)  { outmem[i] = P.at(i,0); }
-        }
+      const unwrap<typename Proxy<T1>::stored_type> tmp(P.Q);
+      
+      arrayops::copy(out.memptr(), tmp.M.memptr(), n_elem);
       }
     else
       {
-      for(uword row=0; row < n_rows; ++row)
-        {
-        uword i,j;
-        
-        for(i=0, j=1; j < n_cols; i+=2, j+=2)
-          {
-          const eT tmp_i = P.at(row,i);
-          const eT tmp_j = P.at(row,j);
-          
-          *outmem = tmp_i; outmem++;
-          *outmem = tmp_j; outmem++;
-          }
-        
-        if(i < n_cols)
-          {
-          *outmem = P.at(row,i); outmem++;
-          }
-        }
+      for(uword i=0; i < n_elem; ++i)  { outmem[i] = P.at(i,0); }
       }
     }
-  else  // we have aliasing
+  else
     {
-    arma_extra_debug_print("op_vectorise_row::apply(): aliasing detected");
-    
-    Mat<eT> tmp;
-    
-    op_vectorise_row::apply_proxy(tmp, P);
-    
-    out.steal_mem(tmp);
+    for(uword row=0; row < n_rows; ++row)
+      {
+      uword i,j;
+      
+      for(i=0, j=1; j < n_cols; i+=2, j+=2)
+        {
+        const eT tmp_i = P.at(row,i);
+        const eT tmp_j = P.at(row,j);
+        
+        *outmem = tmp_i; outmem++;
+        *outmem = tmp_j; outmem++;
+        }
+      
+      if(i < n_cols)
+        {
+        *outmem = P.at(row,i); outmem++;
+        }
+      }
     }
   }
 
@@ -306,7 +331,7 @@ op_vectorise_all::apply(Mat<typename T1::elem_type>& out, const Op<T1,op_vectori
 template<typename T1>
 inline
 void
-op_vectorise_cube_col::apply(Mat<typename T1::elem_type>& out, const BaseCube<typename T1::elem_type, T1>& in)
+op_vectorise_cube_col::apply(Mat<typename T1::elem_type>& out, const CubeToMatOp<T1, op_vectorise_cube_col>& in)
   {
   arma_extra_debug_sigprint();
   
@@ -314,13 +339,18 @@ op_vectorise_cube_col::apply(Mat<typename T1::elem_type>& out, const BaseCube<ty
   
   if(is_same_type< T1, subview_cube<eT> >::yes)
     {
-    op_vectorise_cube_col::apply_subview(out, reinterpret_cast< const subview_cube<eT>& >(in.get_ref()));
+    op_vectorise_cube_col::apply_subview(out, reinterpret_cast< const subview_cube<eT>& >(in.m));
     }
   else
     {
-    const ProxyCube<T1> P(in.get_ref());
-    
-    op_vectorise_cube_col::apply_proxy(out, P);
+    if(is_Cube<T1>::value || (arma_config::openmp && ProxyCube<T1>::use_mp))
+      {
+      op_vectorise_cube_col::apply_unwrap(out, in.m);
+      }
+    else
+      {
+      op_vectorise_cube_col::apply_proxy(out, in.m);
+      }
     }
   }
 
@@ -333,86 +363,101 @@ op_vectorise_cube_col::apply_subview(Mat<eT>& out, const subview_cube<eT>& sv)
   {
   arma_extra_debug_sigprint();
   
-  const uword sv_n_rows   = sv.n_rows;
-  const uword sv_n_cols   = sv.n_cols;
-  const uword sv_n_slices = sv.n_slices;
+  const uword sv_nr = sv.n_rows;
+  const uword sv_nc = sv.n_cols;
+  const uword sv_ns = sv.n_slices;
   
   out.set_size(sv.n_elem, 1);
   
   eT* out_mem = out.memptr();
   
-  for(uword slice=0; slice < sv_n_slices; ++slice)
-  for(uword   col=0;   col < sv_n_cols;   ++col  )
+  for(uword s=0; s < sv_ns; ++s)
+  for(uword c=0; c < sv_nc; ++c)
     {
-    arrayops::copy(out_mem, sv.slice_colptr(slice,col), sv_n_rows);
+    arrayops::copy(out_mem, sv.slice_colptr(s,c), sv_nr);
     
-    out_mem += sv_n_rows;
+    out_mem += sv_nr;
     }
   }
-  
-  
-  
+
+
+
 template<typename T1>
 inline
 void
-op_vectorise_cube_col::apply_proxy(Mat<typename T1::elem_type>& out, const ProxyCube<T1>& P)
+op_vectorise_cube_col::apply_unwrap(Mat<typename T1::elem_type>& out, const T1& expr)
+  {
+  arma_extra_debug_sigprint();
+  
+  const unwrap_cube<T1> U(expr);
+  
+  out.set_size(U.M.n_elem, 1);
+  
+  arrayops::copy(out.memptr(), U.M.memptr(), U.M.n_elem);
+  }
+
+
+
+template<typename T1>
+inline
+void
+op_vectorise_cube_col::apply_proxy(Mat<typename T1::elem_type>& out, const T1& expr)
   {
   arma_extra_debug_sigprint();
   
   typedef typename T1::elem_type eT;
   
+  const ProxyCube<T1> P(expr);
+  
+  if(is_Cube<typename ProxyCube<T1>::stored_type>::value)
+    {
+    op_vectorise_cube_col::apply_unwrap(out, P.Q);
+    
+    return;
+    }
+  
   const uword N = P.get_n_elem();
   
   out.set_size(N, 1);
   
-  if(is_Cube<typename ProxyCube<T1>::stored_type>::value == true)
+  eT* outmem = out.memptr();
+  
+  if(ProxyCube<T1>::use_at == false)
     {
-    const unwrap_cube<typename ProxyCube<T1>::stored_type> tmp(P.Q);
+    typename ProxyCube<T1>::ea_type A = P.get_ea();
     
-    arrayops::copy(out.memptr(), tmp.M.memptr(), N);
+    uword i,j;
+    
+    for(i=0, j=1; j < N; i+=2, j+=2)
+      {
+      const eT tmp_i = A[i];
+      const eT tmp_j = A[j];
+      
+      outmem[i] = tmp_i;
+      outmem[j] = tmp_j;
+      }
+    
+    if(i < N)
+      {
+      outmem[i] = A[i];
+      }
     }
   else
     {
-    eT* outmem = out.memptr();
+    const uword nr = P.get_n_rows();
+    const uword nc = P.get_n_cols();
+    const uword ns = P.get_n_slices();
     
-    if(ProxyCube<T1>::use_at == false)
+    for(uword s=0; s < ns; ++s)
+    for(uword c=0; c < nc; ++c)
+    for(uword r=0; r < nr; ++r)
       {
-      typename ProxyCube<T1>::ea_type A = P.get_ea();
-      
-      uword i,j;
-      
-      for(i=0, j=1; j < N; i+=2, j+=2)
-        {
-        const eT tmp_i = A[i];
-        const eT tmp_j = A[j];
-        
-        outmem[i] = tmp_i;
-        outmem[j] = tmp_j;
-        }
-      
-      if(i < N)
-        {
-        outmem[i] = A[i];
-        }
-      }
-    else
-      {
-      const uword n_rows   = P.get_n_rows();
-      const uword n_cols   = P.get_n_cols();
-      const uword n_slices = P.get_n_slices();
-      
-      for(uword slice=0; slice < n_slices; ++slice)
-      for(uword   col=0;   col < n_cols;   ++col  )
-      for(uword   row=0;   row < n_rows;   ++row  )
-        {
-        *outmem = P.at(row,col,slice);
-        outmem++;
-        }
+      *outmem = P.at(r,c,s);
+      outmem++;
       }
     }
   }
 
 
 
-
 //! @}
